package dareka.processor.impl;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.net.Socket;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Map;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.Main;
import dareka.NLConfig;
import dareka.SocketWrapper;
import dareka.common.Config;
import dareka.common.FileUtil;
import dareka.common.Pair;
import dareka.common.Logger;
import dareka.extensions.Rewriter;
import dareka.processor.HttpHeader;
import dareka.processor.HttpRequestHeader;
import dareka.processor.HttpResponseHeader;
import dareka.processor.HttpUtil;
import dareka.processor.LocalFileResource;
import dareka.processor.Processor;
import dareka.processor.Resource;
import dareka.processor.StringResource;
import dareka.processor.URLResource;

public class RewriterProcessor implements Processor {
    private static final String[] SUPPORTED_METHODS = new String[] {
        HttpHeader.GET, HttpHeader.POST };
    
    protected Config config;
    protected Vector<Rewriter> rewriterList = new Vector<Rewriter>();
    private Rewriter getFlvRewriter;
    private EasyRewriter easyRewriter;
    
    public RewriterProcessor(Config config) {
        this.config = config;
        
        // VXepȈՃtB^o^
        addRewriter(EasyRewriter.getInstance_sys());
        
        // ftHgRewritero^
        addRewriter(new WatchRewriter());
        addRewriter(new SearchRewriter());
        addRewriter(getFlvRewriter = new GetFlvRewriter());
        addRewriter(new ThumbWatchRewriter());
    }
    
// UserFilter̒ǉp
    public void addUserFilter() {
        // ȈՃtB^(SP1.11)
        addRewriter(easyRewriter = EasyRewriter.getInstance());
    }

    public void addRewriter(Rewriter rewriter) {
        rewriterList.add(rewriter);
    }
    
    // Processor interface
    public String[] getSupportedMethods() {
        return SUPPORTED_METHODS;
    }
    
// ΏۂׂĂURLɊg
    private static final Pattern SUPPORTED_PATTERN = Pattern.compile(".+");
    private static final Pattern NICOVIDEO_PATTERN = Pattern.compile(
            "https?://[^/]+\\.(?:nicovideo|nimg|simg)\\.jp/");

    public Pattern getSupportedURLAsPattern() {
        return SUPPORTED_PATTERN;
    }

    public String getSupportedURLAsString() {
        return null;
    }

    private static final Pattern CHARSET_PATTERN = Pattern.compile(
            "(?i)charset\\s*=\\s*([0-9a-z\\-_]+)");
    private static final Pattern HTML_CHARSET_PATTERN = Pattern.compile(
            "(?i)<meta[^>]*charset['\"\\s]*=['\"\\s]*([0-9a-z\\-_]+)");
    
    public Resource onRequest(HttpRequestHeader requestHeader, Socket browser)
            throws IOException {
        String uri = requestHeader.getURI();
        InputStream in = browser.getInputStream();
        
        if (isDebugURL(uri)) {
            Logger.info("RewriterURL = " + uri);
        }
        
// POST̏ɑΉ
        String postString = null;
        ArrayList<Pair<Rewriter, Matcher>> postRW = null;
        if (HttpHeader.POST.equals(requestHeader.getMethod())) {
            String postURI = uri.replaceFirst("^(https?)://", "$1://POST/");
            postRW = getMatchedRewriter(postURI);
            long contentLength = requestHeader.getContentLength();
            if (postRW.size() > 0 &&
                    contentLength <= Long.getLong("postSizeMax", 16*1024*1024)) {
                postString = getPostString(requestHeader, browser, "UTF-8");
                String replaced = doRewriter(postURI, postString,
                        requestHeader, getDummyResponseHeader(), postRW, null);
                if (postString != replaced) {
                    byte[] postBuf = replaced.getBytes("UTF-8");
                    requestHeader.setContentLength(postBuf.length);
                    in = new ByteArrayInputStream(postBuf);
                }
                if (Boolean.getBoolean("debugRewriterOriginal")) {
                    debugOutputBody(uri, "PostOrgBody", postString);
                }
                debugOutputBody(uri, "RequestBody", replaced);
            }
        }
        
// getflv́Apost̓eRewriterɕKvȂ̂œʈ
// getflv̋AĂȂIDcontentID킩Ȃ̂ŁA
// RewriterɃp[^^ׂURIs
// (Iւ̃ANZXRequestHeaderȂ̂Ō̂܂)
        if (getFlvRewriter != null && 
                postString != null &&
                getFlvRewriter.getRewriterSupportedURLAsPattern().matcher(uri).lookingAt()) {
            uri += "&" + postString;
        }
        
        ArrayList<Pair<Rewriter, Matcher>> rewriters = getMatchedRewriter(uri);
        int rwSize = rewriters.size();
        if (rwSize == 0 && (postRW == null || postRW.isEmpty())) {
            // uΏۂłȂ΂̂܂ܕԂ
            return Resource.get(Resource.Type.URL, requestHeader.getURI());
        }
        
        // uΏۂΈU_E[hRewriterŏ
        requestHeader.removeHopByHopHeaders();
// uꍇ͉𓀂łȂEncoding͍폜
        String acceptEncoding = requestHeader.getMessageHeader(HttpHeader.ACCEPT_ENCODING);
        if (rwSize > 0 && acceptEncoding != null) {
            acceptEncoding = acceptEncoding.toLowerCase().replaceAll("( *,)?(bzip2|sdch)", "");
            acceptEncoding = acceptEncoding.replaceFirst("^ *," ,"");
            requestHeader.setMessageHeader(HttpHeader.ACCEPT_ENCODING, acceptEncoding);
        }
        
        // wb_MāABodyM邩f
        URLResource r = new URLResource(requestHeader.getURI());
        HttpResponseHeader responseHeader = r.getResponseHeader(in, requestHeader);
        if (responseHeader == null) {
            Logger.warning("failed to rewrite: " + uri + " (no responseHeader)");
            return r;
        }
        
        // nlFiltergɑΉ邽EasyRewriterOŏ
        ArrayList<UserFilter> userFilters = null;
        if (rwSize > 0 && rewriters.get(rwSize - 1).first == easyRewriter) {
            userFilters = easyRewriter.getMatchedUserFilter(
                    uri, requestHeader, responseHeader);
            if (userFilters.isEmpty()) {
                // KptB^EasyRewriter폜
                rewriters.remove(rwSize - 1);
            }
        }
        
        // URL.php͈ꉞΏ
        // ̑̃}b`͊OɈړ
        String contentType = responseHeader.getMessageHeader(HttpHeader.CONTENT_TYPE);
        if (rwSize > 0 &&
            (responseHeader.getStatusCode() == 200 ||
             responseHeader.getStatusCode() == 403 ||
             responseHeader.getStatusCode() == 404 ||
             responseHeader.getStatusCode() == 503)
            && (uri.contains(".php") || matchContentType(contentType, userFilters)))
        {
            // Body̎擾
            byte[] bcontent = r.getResponseBody();
            if (bcontent == null) {
                Logger.warning("failed to rewrite: " + uri + " (no responseBody)");
                return r;
            }
            if (responseHeader.getMessageHeader(HttpHeader.CONTENT_ENCODING) != null) {
                Logger.warning("failed to rewrite: " + uri + " (cannot decode)");
                return r;
            }
            responseHeader.removeMessageHeader("Vary");
            responseHeader.removeMessageHeader("Accept-Ranges");
            responseHeader.removeHopByHopHeaders();
            
            // nlp[^̃Rs[(rc2.03)
            for (Map.Entry<String, String> e : requestHeader.getParameters().entrySet()) {
                responseHeader.setParameter(e.getKey(), e.getValue());
            }
            bcontent = doRewriter(uri, bcontent,
                    requestHeader, responseHeader, rewriters, userFilters);
            return StringResource.getRawResource(responseHeader, bcontent);
        }
        else {
            // wb_ႤƔ肵͉̂ςȂőM
            responseHeader.removeMessageHeader("Vary");
            responseHeader.removeMessageHeader("Accept-Ranges");
            Logger.debugWithThread("rewriter content-type: " + contentType);
        }
        
        return r;
    }
    
    private boolean isDebugURL(String url) {
        if (Boolean.getBoolean("debugRewriter")) {
            Matcher m = NLConfig.getMatcher("debugRewriterURLPattern", url);
            return m == null || m.find();
        }
        return false;
    }
    
    private void debugOutputBody(String url, String prefix, String content) {
        if (isDebugURL(url)) {
            int length = Integer.getInteger("debugRewriterMaxLength", 0);
            if (length > 0 && content.length() > length) {
                content = content.substring(0, length);
            }
            Logger.info(prefix + ": " + content);
        }
    }
    
    /**
     * NGXgPOST̏ꍇɁA͓̓ew蕶ZbgŎ擾ԂB
     * <p>͕Zbgnullw肵ꍇ ISO-8859-1 Ŏ擾āAXɓ͂
     * <b>application/x-www-form-urlencoded</b>ł UTF-8 ŃfR[hsB
     * <p>nl{̑œ͂obt@OĎ擾ȌԂɖ߂̂ŁA{\bh
     * {@link Processor#onRequest(HttpRequestHeader, Socket)}
     * ĂяoĂAȍ~̏ōĂѓ͓e擾łB
     * 
     * @param requestHeader NGXgwb_
     * @param browser \Pbg
     * @param charsetName ͕ZbgAnullȂ"ISO-8859-1"łfR[hL
     * @return ̂ꂩ̒lԂ
     * <ul>
     * <li>NGXgPOSTȂ͓ew蕶ZbgŎ擾
     * <li>NGXgPOSTł͖ꍇnull
     * <li>͂Content-Lengths̏ꍇ͋󕶎
     * </ul>
     * @throws IOException o̓G[ꍇ
     * @since NicoCache_nl+110522mod
     */
    public String getPostString(HttpRequestHeader requestHeader, Socket browser,
            String charsetName) throws IOException {
        if (!HttpHeader.POST.equals(requestHeader.getMethod())) {
            return null;
        }
        long contentLength = requestHeader.getContentLength();
        if (contentLength <= 0L || Integer.MAX_VALUE < contentLength) {
            Logger.debugWithThread("invalid POST contentLength: " + contentLength);
            return "";
        }
        byte[] buf = new byte[(int) contentLength];
        int pos = 0;
        while (pos < buf.length) {
            int len = buf.length - pos;
            pos += browser.getInputStream().read(buf, pos, len);
        }
        if (browser instanceof SocketWrapper) {
            ((SocketWrapper) browser).resetInputStreamBuffer();
        }
        if (charsetName == null) {
            String postString = new String(buf, "ISO-8859-1");
            if ("application/x-www-form-urlencoded".equalsIgnoreCase(
                    requestHeader.getMessageHeader(HttpHeader.CONTENT_TYPE))) {
                postString = URLDecoder.decode(postString, "UTF-8");
            }
            return postString;
        }
        return new String(buf, charsetName);
    }
    
    // Content-TypeRewritersĂǂ𔻒肷
    private boolean matchContentType(
            String contentType, ArrayList<UserFilter> userFilters) {
        if (contentType == null) {
            return true; // Content-Typeꍇ͓eȂ̂ŋ
        }
        if (NLConfig.find("rewriterContentType", contentType)) {
            return true;
        }
        // nlFilterContentType=w肪Ȃ炻`FbN
        if (userFilters != null && userFilters.size() > 0) {
            for (UserFilter u : userFilters) {
                if (u.contentType != null) {
                    return true; // Ƀ}b`ςȂ̂ő݂Ȃ疳ŋ
                }
            }
        }
        return false;
    }
    
    /**
     * URLɃ}b`RewriterԂBp[^͎菜ă}b`OsB
     * ̊gq(rewriterIgnoreExt)ŏIURL̏ꍇ͏Ƀ}b`ȂB
     * 
     * @param url }b`ΏۂURL
     * @return }b`Rewriter̃XgA}b`Ȃ΋󃊃Xg
     * @since NicoCache_nl+101219mod
     */
    ArrayList<Pair<Rewriter, Matcher>> getMatchedRewriter(String url) {
        ArrayList<Pair<Rewriter, Matcher>> matched =
            new ArrayList<Pair<Rewriter, Matcher>>(rewriterList.size());
        if (!isIgnoreExt(url)) {
            Matcher m;
            if (LocalDirProcessor.isSupportedURL(url)) {
                // [J̏ꍇnlFilter̂ݑΏ
                m = easyRewriter.getRewriterSupportedURLAsPattern().matcher(url);
                if (m.lookingAt()) {
                    matched.add(new Pair<Rewriter, Matcher>(easyRewriter, m));
                    return matched;
                }
            }
            for (Rewriter rw : rewriterList) {
                m = rw.getRewriterSupportedURLAsPattern().matcher(url);
                if (m.lookingAt()) {
                    matched.add(new Pair<Rewriter, Matcher>(rw, m));
                }
            }
        }
        return matched;
    }
    
    private boolean isIgnoreExt(String url) {
        int end = url.indexOf('?'); // p[^͖
        if (end < 0) end = url.length();
        for (int i = end - 1; i >= 0; i--) {
            if (url.charAt(i) == '.') {
                String ext = url.substring(i + 1, end);
                if (NLConfig.find("rewriterExtIgnore", ext)) {
                    return true;
                }
                break;
            } else if (url.charAt(i) == '/') {
                break;
            }
        }
        return false;
    }
    
    // Rewriter(Recp)
    private byte[] doRewriter(String url, byte[] bcontent,
            HttpRequestHeader requestHeader, HttpResponseHeader responseHeader,
            ArrayList<Pair<Rewriter, Matcher>> rewriters,
            ArrayList<UserFilter> userFilters) throws IOException {
        String original = null;
        String charset = getDefaultCharset(url);
        String contentType = responseHeader.getMessageHeader(HttpHeader.CONTENT_TYPE);
        
        if (contentType != null) {
            Matcher m = CHARSET_PATTERN.matcher(contentType);
            boolean found = m.find();
            if (found && Charset.isSupported(m.group(1))) {
                charset = m.group(1);
            } else if (contentType.startsWith("text/html")) {
                original = new String(bcontent, "ISO-8859-1");
                m = HTML_CHARSET_PATTERN.matcher(original);
                if (m.find() && Charset.isSupported(m.group(1))) {
                    charset = m.group(1);
                }
                if (!"ISO-8859-1".equalsIgnoreCase(charset)) {
                    original = null;
                }
            }
            if (!found && charset != null) {
                contentType += "; charset=" + charset;
            }
        }
        
        Logger.debugWithThread("rewriter content-type: " + contentType);
        if (charset == null) {
            Logger.debugWithThread("no rewriter content");
            return bcontent;
        } else if (original == null) {
            original = new String(bcontent, charset);
        }
        String replaced = doRewriter(url, original,
                requestHeader, responseHeader, rewriters, userFilters);
        if (original != replaced) {
            bcontent = replaced.getBytes(charset);
            // uꍇLast-ModifiedăuEUϋɓI
            // LbVȂ悤ɂ(Cache-Control܂ł͂Ȃ)
            responseHeader.removeMessageHeader(HttpHeader.LAST_MODIFIED);
        }
        if (Boolean.getBoolean("debugRewriterOriginal")) {
            debugOutputBody(url, "OriginalBody", original);
        }
        debugOutputBody(url, "ResponseBody", replaced);
        
        return bcontent;
    }
    
    private String getDefaultCharset(String url) {
        // jRjRȂUTF-8Ōߑł
        if (NICOVIDEO_PATTERN.matcher(url).lookingAt()) {
            return "UTF-8";
        }
        String charset = System.getProperty("rewriterDefaultCharset", "");
        if (charset.length() > 0 || Charset.isSupported(charset)) {
            return charset;
        }
        return null;
    }
    
    // Rewriter(p)
    private String doRewriter(String url, String content,
            HttpRequestHeader requestHeader, HttpResponseHeader responseHeader,
            ArrayList<Pair<Rewriter, Matcher>> rewriters,
            ArrayList<UserFilter> userFilters) throws IOException {
        if (EasyRewriter.checkDebugMode()) {
            Logger.info("[Debug]Filter Processing start. URL: " + url);
        }
        long startTime = System.nanoTime();
        if (rewriters == null) {
            rewriters = getMatchedRewriter(url);
        }
        String original = content;
        for (Pair<Rewriter, Matcher> entry : rewriters) {
            if (entry.first != easyRewriter) {
                content = entry.first.onMatch(entry.second, responseHeader, content);
            } else {
                // EasyRewriter𒼐ڏ
                if (userFilters == null) {
                    userFilters = easyRewriter.getMatchedUserFilter(
                            url, requestHeader, responseHeader);
                }
                content = easyRewriter.applyUserFilter(
                        url, content, requestHeader, responseHeader, userFilters);
            }
        }
        if (EasyRewriter.checkDebugMode() || Boolean.getBoolean("dareka.debug")) {
            long elapsedTime = System.nanoTime() - startTime;
            if (EasyRewriter.checkDebugMode()) {
                Logger.info("[Debug]Filter Processing end. time: %,dms",
                        (int)(elapsedTime / 1000000));
            }
            String result = original != content ? " replaced" : "";
            Logger.debugWithThread(String.format(
                    "rewriter elapsed time: %,dns%s", elapsedTime, result));
        }
        return content;
    }
    
    private HttpResponseHeader getDummyResponseHeader() throws IOException {
        return new HttpResponseHeader(
                "HTTP/1.1 200 OK\r\n" +
                "Content-Type: text/plain\r\n\r\n");
    }
    
    /**
     * RewriterŒuB
     * <p>
     * NGXgwb_уX|Xwb_nlFilteru(RequestHeader,
     * ContentType )邽߂ɕKvȂ̂ŁAȂׂw肷邱ƁB
     * ȂAX|Xwb_nullnꍇ"text/plain"ƂĈB
     * 
     * @param url RewriterΏۂURL
     * @param content uΏۂ̕
     * @param requestHeader NGXgwb_AsvȂnull
     * @param responseHeader X|Xwb_AsvȂnull
     * @return uAuΏۂŖcontent̂܂ܕԂ
     * @since NicoCache_nl+101219mod
     */
    public String stringRewriter(String url, String content,
            HttpRequestHeader requestHeader, HttpResponseHeader responseHeader)
    throws IOException {
        ArrayList<Pair<Rewriter, Matcher>> rewriters = getMatchedRewriter(url);
        if (rewriters.size() > 0) {
            if (responseHeader == null) {
                responseHeader = getDummyResponseHeader();
            }
            content = doRewriter(url, content,
                    requestHeader, responseHeader, rewriters, null);
        }
        return content;
    }
    
    /**
     * nlFilter̒u(RequireHeader, ContentType )ɑΉłȂ̂
     * HttpRequestHeader, HttpResponseHeader ̂łgĂB
     * @see #stringRewriter(String, String, HttpRequestHeader, HttpResponseHeader)
     */
    @Deprecated
    public String stringRewriter(String url, String content) throws IOException {
        return stringRewriter(url, content, null, null);
    }
    
    /**
     * Rec(oCgz)RewriterŒuBR[h͓K؂ɏB
     * 
     * @param url RewriterΏۂURL
     * @param bcontent uΏۂ̃Rec
     * @param requestHeader NGXgwb_
     * @param responseHeader X|Xwb_
     * @return uRecAuΏۂŖbcontent̂܂ܕԂ
     * @see #stringRewriter(String, String, HttpRequestHeader, HttpResponseHeader)
     * @since NicoCache_nl+101219mod
     */
    public byte[] contentRewriter(String url, byte[] bcontent,
            HttpRequestHeader requestHeader, HttpResponseHeader responseHeader) {
        if (url == null && requestHeader != null) {
            url = requestHeader.getURI();
        }
        return contentRewriter(url, bcontent,
                requestHeader, responseHeader, getMatchedRewriter(url), null);
    }
    
    private byte[] contentRewriter(String url, byte[] bcontent,
            HttpRequestHeader requestHeader, HttpResponseHeader responseHeader,
            ArrayList<Pair<Rewriter, Matcher>> rewriters,
            ArrayList<UserFilter> userFilters) {
        if (rewriters != null && bcontent != null) {
            try {
                return doRewriter(url, decode(bcontent, responseHeader),
                        requestHeader, responseHeader, rewriters, userFilters);
            } catch (IOException e) {
                Logger.error(e);
            }
        }
        return bcontent;
    }
    
    private byte[] decode(byte[] bcontent, HttpResponseHeader responseHeader)
            throws IOException {
        if (responseHeader != null) {
            String enc = responseHeader.getMessageHeader(HttpHeader.CONTENT_ENCODING);
            if (enc != null) {
                ByteArrayOutputStream bout = new ByteArrayOutputStream();
                FileUtil.copy(HttpUtil.getDecodedInputStream(bcontent, enc), bout);
                bcontent = bout.toByteArray();
                responseHeader.removeMessageHeader(HttpHeader.CONTENT_ENCODING);
            }
        }
        return bcontent;
    }
    
    /**
     * [Jt@CRewriterŒu(localRewriter=truȅꍇ̂)B
     * e\[XƂĕԂB
     * 
     * @param url RewriterΏۂURL
     * @param file uΏۂ̃[Jt@C
     * @param requestHeader NGXgwb_
     * @return u\[X
     * @see #stringRewriter(String, String, HttpRequestHeader, HttpResponseHeader)
     * @since NicoCache_nl+101219mod
     */
    public Resource localRewriter(String url, File file,
            HttpRequestHeader requestHeader) throws IOException {
        if (file == null || file.isDirectory()) {
            return StringResource.getNotFound();
        }
        LocalFileResource r = new LocalFileResource(file);
        if (!Boolean.getBoolean("localRewriter")) {
            return r;
        }
        RewriterProcessor rewriter = Main.getRewriterProcessor();
        ArrayList<Pair<Rewriter, Matcher>> rewriters = rewriter.getMatchedRewriter(url);
        if (rewriters.size() == 0) {
            return r;
        }
//      // K邽߂If-Modified-Since͍폜
//      requestHeader.removeMessageHeader(HttpHeader.IF_MODIFIED_SINCE);
        HttpResponseHeader responseHeader = r.getResponseHeader(requestHeader, true);
        ArrayList<UserFilter> userFilters =
            easyRewriter.getMatchedUserFilter(url, requestHeader, responseHeader);
        if (userFilters.size() == 0) {
            return r;
        }
        // K邽߂If-Modified-Since폜čĎ擾
        requestHeader.removeMessageHeader(HttpHeader.IF_MODIFIED_SINCE);
        responseHeader = r.getResponseHeader(requestHeader, true);
        byte[] bcontent = rewriter.contentRewriter(url, r.getResponseBody(),
                requestHeader, responseHeader, rewriters, userFilters);
        return StringResource.getRawResource(responseHeader, bcontent);
    }
    
    /**
     * file = new File(path)
     * @param path [Jt@C̃pX
     * @see #localRewriter(String, File, HttpRequestHeader)
     */
    public Resource localRewriter(String uri, String path,
            HttpRequestHeader requestHeader) throws IOException {
        return localRewriter(uri, new File(path), requestHeader);
    }
}
