package dareka.processor.impl;

import java.io.FileReader;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.Main;
import dareka.NLConfig;
import dareka.common.LRUMap;
import dareka.common.Logger;
import dareka.common.TextUtil;
import dareka.common.json.*;
import dareka.processor.HttpHeader;
import dareka.processor.HttpRequestHeader;
import dareka.processor.HttpResponseHeader;

/**
 * [nl] watch y[Wp[XăvC[ɐݒ肷ϐ⓮̒l擾NX
 * <ul>
 * <li>h̏ꍇ so.addVariable/var Video 擾
 * <li>GINZȀꍇ watchAPIDataContainer 擾
 * <li>HTML5̏ꍇ data-api-data 擾
 * <li>uIɃjR HTML5 x[XɂȂꍇŋzv
 *     łDȂقǐ_ˁ[
 * </ul>
 * @since NicoCache_nl+120609mod
 */
public class WatchVars {
    // h
    private static final Pattern SO_ADDVARIABLE_PATTERN = Pattern.compile(
            "so\\.addVariable\\(\\s*(\".+?\")\\s*,\\s*\"(.*?)\"\\);");
    private static final Pattern VAR_VIDEO_PATTERN = Pattern.compile(
            "<!--\\s+var Video = (\\{[\\s\\S]+?\\});\\s+(?:--></script>|var)");
    
    // Zero-GINZA
    private static final Pattern WATCH_API_DATA_CONTAINER_PATTERN = Pattern.compile(
            "<div id=\"watchAPIDataContainer\"[^>]*>([\\s\\S]+?)</div>");
    private static final Pattern PLAYLIST_DATA_CONTAINER_PATTERN = Pattern.compile(
            "<div id=\"playlistDataContainer\"[^>]*>([\\s\\S]+?)</div>");
    
    // HTML5
    private static final Pattern DATA_API_DATA_CONTAINER_PATTERN = Pattern.compile(
            "data-api-data=\"([^\"]+)\"");
    
    
    private static final Pattern WATCH_ID_PATTERN = Pattern.compile(
            "([a-z]{2})?(\\d+)");
    private static final Pattern WATCH_ECONOMY_MODE_PATTERN = Pattern.compile(
            "\\?(eco|lo)=1");
    
    private static final Map<String, WatchVars> cache =
            Collections.synchronizedMap(new LRUMap<String, WatchVars>());
    
    /**
     * WatchVars IuWFNg擾
     * <p>
     * ^ꂽ watch y[W̓e WatchVars IuWFNg𐶐ĕԂB
     * IDƃ^Cg擾łꍇ {@linkplain NicoIdInfoCache} ɋL^B
     * XɐIuWFNgLbVB
     * <p>
     *  https?:// n܂镶A12ȉ̕^ꍇA
     * ^ꂽɓID擾A̓IDɑΉ WatchVars
     * IuWFNgLbVɎcĂȂ炻ԂB
     * <p>
     * ɍŐV̏擾ꍇ́Awatch y[W̓en WatchVars
     * IuWFNg𐶐KvLB
     * ȂAIuWFNg̐ɂ͂x̎ԂKvȂ̂ŁARewriter
     * ̌ĂяoŎO watch y[WʂĂ鎖łꍇA
     * LbV𗘗pptH[}XIɗLłB
     * 
     * @param content watch y[W̓eA URL or ID or XbhID
     * @return WatchVars IuWFNgALbVɎcĂȂꍇ̂ null
     */
    public static WatchVars get(String content) {
        if (content == null) {
            return null;
        }
        if (content.length() <= 12 || content.matches("https?://.*")) {
            int pos = content.lastIndexOf('/');
            if (pos > 0) {
                content = content.substring(pos + 1);
            }
            Matcher m = WATCH_ID_PATTERN.matcher(content);
            if (!m.find()) {
                return null;
            }
            String videoId = m.group();
            if (TextUtil.isThreadId(videoId)) {
                videoId = NLShared.INSTANCE.thread2smid(videoId);
                if (videoId == null) {
                    return null;
                }
            }
            return cache.get(videoId);
        }
        return new WatchVars(content);
    }
    
    private String content, data;
    private Matcher dataMatcher;
    private boolean harajuku = false;
    private boolean ginza = false;
    private boolean rawJson = false;
    private JsonObject json;
    private JsonObject flashvars;
    private JsonObject dmcInfo;
    private boolean isDmc;
    private String videoId, videoTitle;
    private boolean isPeakTime, isPremium;
    private long duration;
    private String recipe_id = null;
    private List<String> videos = null;
    private List<String> audios = null;
    private Map<String, Boolean> qualityVideos = null;
    private Map<String, Boolean> qualityAudios = null;
    private HttpRequestHeader requestHeader;
    
    private WatchVars(String content) {
        this.content = content;
        dataMatcher = DATA_API_DATA_CONTAINER_PATTERN.matcher(content);
        if (!dataMatcher.find()) {
            dataMatcher = WATCH_API_DATA_CONTAINER_PATTERN.matcher(content);
            if (dataMatcher.find()) {
                ginza = true;
            } else {
                dataMatcher = null;
            }
        }
        if (dataMatcher != null) {
            data = TextUtil.unescapeHTML(dataMatcher.group(1));
            json = Json.parseObject(data);
        } else if (content.startsWith("{")) {
            data = content;
            json = Json.parseObject(data);
            rawJson = true;
        } else {
            data = content;
            json = harajuku2json(data);
            if (json != null) {
                harajuku = true;
            }
        }
        if (json != null) {
            if (dataMatcher == null && !json.get("flashvars").isNull()) {
                ginza = true;
            }
            if (!ginza && !harajuku) {
                dmcInfo = json.getObject("video", "dmcInfo");
                isDmc = dmcInfo != null;
                videoId = json.getString("video", "id");
                if (videoId != null) {
                    // titleoriginalTitlëႢ͉낤
                    // originalDescriptionɂ̓NHTMLĂ̂
                    // titlelƂoriginalTitlegׂ
                    videoTitle = json.getString("video", "originalTitle");
                    NicoCachingTitleRetriever.putTitleCache(videoId, videoTitle);
                    
                    isPeakTime = json.getBoolean("context", "isPeakTime");
                    isPremium = json.getBoolean("viewer", "isPremium");
                    
                    if (isDmc) {
                        recipe_id = dmcInfo.getString("session_api", "recipe_id");
                        
                        videos = getArrayOfString(dmcInfo.getArray("session_api", "videos"));
                        audios = getArrayOfString(dmcInfo.getArray("session_api", "audios"));
                        
                        qualityVideos = getQualityAvailability(dmcInfo.getArray("quality", "videos"));
                        qualityAudios = getQualityAvailability(dmcInfo.getArray("quality", "audios"));
                    }
                    
                    JsonValue durationValue = null;
                    if (dmcInfo != null) {
                        durationValue = dmcInfo.get("video", "length_seconds");
                    }
                    if (durationValue == null || durationValue.isNull()) {
                        durationValue = json.get("video", "duration");
                    }
                    if (durationValue instanceof JsonNumber) {
                        duration = ((JsonNumber) durationValue).getLong();
                    }
                }
            } else if (ginza) {
                JsonValue v = json.get("flashvars");
                if (v instanceof JsonObject) {
                    flashvars = (JsonObject) v;
                }
                isDmc = json.getBoolean("flashvars", "isDmc");
                
                videoId = json.getString("flashvars", "videoId");
                if (videoId != null) {
                    videoTitle = json.getString("flashvars", "videoTitle");
                    NicoCachingTitleRetriever.putTitleCache(videoId, videoTitle);

                    isPeakTime = json.getBoolean("flashvars", "isPeakTime");
                    isPremium = json.getBoolean("viewerInfo", "isPremium");
                    
                    if (isDmc) {
                        String dmcInfoString = json.getString("flashvars", "dmcInfo");
                        if (dmcInfoString != null) {
                            dmcInfo = Json.parseObject(decodeURL(dmcInfoString));
                            recipe_id = dmcInfo.getString("session_api", "recipe_id");
                            videos = getArrayOfString(dmcInfo.getArray("session_api", "videos"));
                            audios = getArrayOfString(dmcInfo.getArray("session_api", "audios"));
                        } else {
                            isDmc = false;
                        }
                    }
                    
                    JsonValue durationValue = null;
                    if (dmcInfo != null) {
                        durationValue = dmcInfo.get("video", "length_seconds");
                    }
                    if (durationValue == null || durationValue.isNull()) {
                        durationValue = json.get("videoDetail", "length");
                    }
                    if (durationValue instanceof JsonNumber) {
                        duration = ((JsonNumber) durationValue).getLong();
                    }
                }
            } else if (harajuku) {
                videoId = json.getString("videoId");
                if (videoId != null) {
                    videoTitle = json.getString("videoTitle");
                    NicoCachingTitleRetriever.putTitleCache(videoId, videoTitle);
                    
                    isPeakTime = json.getBoolean("isPeakTime");
                    isPremium = Pattern.matches("[\\S\\s]*var User =.*?isPremium:\\s*true,[\\S\\s]*", content);
                }
            }
            if (videoId != null) {
                if (isDmc && videos != null && audios != null && recipe_id != null) {
                    String dmcType, dmcId;
                    Matcher m = Pattern.compile("nicovideo-(\\w{2})(\\d+)").matcher(recipe_id);
                    if (m.matches()) {
                        dmcType = m.group(1);
                        dmcId = m.group(2);
                    } else {
                        Logger.warning("WatchVars: invalid recipe_id");
                        dmcType = videoId.substring(0, 2);
                        dmcId = videoId.substring(2);
                    }
                    
                    boolean economy = isDmcEconomy();
                    // IDvĂƂ
                    // NicoCachingTitleRetriever.putTitleCacheʂĂ̂
                    // Gg͕K݂
                    if (!dmcId.equals(videoId.substring(2))) {
                        NicoIdInfoCache.getInstance().putOnlyTypeAndId(dmcType, dmcId);
                    }
                    NicoIdInfoCache.Entry info =
                            NicoIdInfoCache.getInstance().get(dmcId);
                    if (qualityVideos != null && qualityAudios != null) {
                        boolean kuLocked = false;
                        boolean low;
                        low = false;
                        for (Map.Entry<String, Boolean> e : qualityVideos.entrySet()) {
                            boolean kuLockedVideo = isKuLockedVideo(e.getKey());
                            info.dmcVideoEconomy.put(e.getKey(), low);
                            info.dmcVideoKuLocked.put(e.getKey(), kuLockedVideo);
                            if (kuLockedVideo) {
                                kuLocked = true;
                            } else {
                                low = true;
                            }
                        }
                        low = false;
                        for (Map.Entry<String, Boolean> e : qualityAudios.entrySet()) {
                            boolean kuLockedAudio = isKuLockedAudio(e.getKey());
                            info.dmcAudioEconomy.put(e.getKey(), low);
                            info.dmcAudioKuLocked.put(e.getKey(), kuLockedAudio);
                            if (kuLockedAudio) {
                                kuLocked = true;
                            } else {
                                low = true;
                            }
                        }
                        info.setKuLocked(kuLocked);
                        
                        // heuristicsGW̌؁D̂D
                        // [kuLocked => isKuLockedHeuristic()] ɐD
                        if (kuLocked && !isKuLockedHeuristic()) {
                            reportKuLockedBug();
                        }
                    } else {
                        for (int i = 0; i < videos.size(); i++) {
                            info.dmcVideoEconomy.put(videos.get(i), i != 0 || economy);
                        }
                        for (int i = 0; i < audios.size(); i++) {
                            info.dmcAudioEconomy.put(audios.get(i), i != 0 || economy);
                        }
                        info.setKuLocked(isKuLockedHeuristic());
                    }
                }
                cache.put(videoId, this);
            }
        } else {
            Logger.warning("WatchVars: JSON not found.");
        }
        if (Boolean.getBoolean("debugWatchVars")) {
            Logger.info("WatchVars: JSON " + this.getJSON());
            Logger.info("WatchVars: INFO " + videoId + " "
                    + (ginza ? "GINZA " : "")
                    + (harajuku ? "Harajuku " : "")
                    + (isDmc ? "dmc" : "classic") + " " + videoTitle);
            Logger.info("WatchVars: INFO2"
                    + (isPremium ? " Premium" : " !Premium")
                    + (isPeakTime ? " PeakTime" : " !PeakTime"));
            if (isDmc) {
                Logger.info("WatchVars: dmcInfo.recipe_id=" + recipe_id);
                String dmcId;
                Matcher m = Pattern.compile("nicovideo-(\\w{2})(\\d+)").matcher(recipe_id);
                if (m.matches()) {
                    dmcId = m.group(2);
                } else {
                    Logger.info("WatchVars: invalid recipe_id");
                    dmcId = videoId.substring(2);
                }
                NicoIdInfoCache.Entry info =
                        NicoIdInfoCache.getInstance().get(dmcId);
                boolean kuLocked = info == null ? false : info.isKuLocked();
                if (qualityVideos != null) {
                    Logger.info("WatchVars: VIDEOS (all)");
                    for (Map.Entry<String, Boolean> e : qualityVideos.entrySet()) {
                        Boolean low = info == null ? null : info.dmcVideoEconomy.get(e.getKey());
                        Logger.info("WatchVars: - " + (e.getValue() ? "" : "!") + e.getKey()
                                + (low == null ? " null" : low ? " low" :
                                        kuLocked && !isKuLockedVideo(e.getKey()) ? " kulow" : "")
                                + (isKuLockedVideo(e.getKey()) ? " ku" : ""));
                    }
                } else if (videos != null) {
                    Logger.info("WatchVars: VIDEOS (availables)");
                    for (String video : videos) {
                        Boolean low = info == null ? null : info.dmcVideoEconomy.get(video);
                        Logger.info("WatchVars: - " + video
                                + (low == null ? " null" : low ? " low" : kuLocked ? " kulow" : ""));
                    }
                }
                if (qualityAudios != null) {
                    Logger.info("WatchVars: AUDIOS (all)");
                    for (Map.Entry<String, Boolean> e : qualityAudios.entrySet()) {
                        Boolean low = info == null ? null : info.dmcAudioEconomy.get(e.getKey());
                        Logger.info("WatchVars: - " + (e.getValue() ? "" : "!") + e.getKey()
                                + (low == null ? " null" : low ? " low" :
                                        kuLocked && !isKuLockedAudio(e.getKey()) ? " kulow" : "")
                                + (isKuLockedAudio(e.getKey()) ? " ku" : ""));
                    }
                } else if (audios != null) {
                    Logger.info("WatchVars: AUDIOS (availables)");
                    for (String audio : audios) {
                        Boolean low = info == null ? null : info.dmcAudioEconomy.get(audio);
                        Logger.info("WatchVars: - " + audio
                                + (low == null ? " null" : low ? " low" : kuLocked ? " kulow" : ""));
                    }
                }
                Logger.info("WatchVars: INFO3"
                        + (isDmcEconomy() ? " DmcEconomy" : " !DmcEconomy")
                        + (kuLocked ? " kuLocked" : " !kuLocked"));
            }
            StringBuilder sb = new StringBuilder("WatchVars: TAGS");
            for (String tag : getTags()) sb.append(" ").append(tag);
            Logger.info(sb.toString());
        }
    }

    // h watch y[W JSON ɂłグ(watchAPI ̃vCXg`)
    private static JsonObject harajuku2json(String contents) {
        boolean found = false;
        StringBuilder sb = new StringBuilder("{");
        
        Matcher m = SO_ADDVARIABLE_PATTERN.matcher(contents);
        while (m.find()) {
            found = true;
            if (sb.length() > 1) {
                sb.append(",");
            }
            sb.append(m.group(1)).append(":");
            
            // ɂȂ烁oɂĕƐlgKv
            String s = TextUtil.escapeJSON(decodeURL(m.group(2)));
            sb.append('"').append(s).append('"');
        }
        if (!found) {
            return null;
        }
        if ((m = VAR_VIDEO_PATTERN.matcher(contents)).find()) {
            // JavaScript  JSON ɕϊ
            String s = m.group(1)
                    .replaceAll("\\s*//[^\"\n]+", "")
                    .replaceAll("\\s+(\\w+):\\s+", "\"$1\":")
                    .replaceAll("\\s*\\[\\s+([\\S]+?)\\s+\\]", "[$1]")
                    .replaceAll("'", "\"");
            sb.append(",\"videoDetail\":").append(s);
        }
        
        return Json.parseObject(sb.append("}").toString());
    }

    private static final Pattern VIDEO_MODE_PATTERN = Pattern.compile("^(\\d+)([pi])$");
    private static final Pattern VIDEO_TYPE_PATTERN = Pattern.compile("^archive_(?:[^_]+)_(\\d+)kbps_([^_]+)$");
    private static final Pattern AUDIO_TYPE_PATTERN = Pattern.compile("^archive_(?:[^_]+)_(\\d+)kbps$");
    
    /**
     * dmc̃GRm~[(q[XeBbN)
     */
    private boolean isDmcEconomy() {
        if (isPremium || !isPeakTime) {
            return false;
        }
        if (videos == null || videos.size() == 0 ||
                audios == null || audios.size() == 0) {
            return false;
        }
        String video = videos.get(0);
        String audio = audios.get(0);
        int videoBitrate = Integer.MAX_VALUE;
        int audioBitrate = Integer.MAX_VALUE;
        Matcher mv = VIDEO_TYPE_PATTERN.matcher(video);
        if (mv.matches()) {
            videoBitrate = Integer.parseInt(mv.group(1));
        }
        Matcher ma = AUDIO_TYPE_PATTERN.matcher(audio);
        if (ma.matches()) {
            audioBitrate = Integer.parseInt(ma.group(1));
        }
        return videoBitrate <= 600 && audioBitrate <= 64;
    }
    
    private boolean isKuLockedHeuristic() {
        if (!isPremium && isPeakTime) {
            return true;
        }
        if (videos == null || videos.size() == 0) {
            return true;
        }
        String video = videos.get(0);
        int bitrate = Integer.MAX_VALUE;
        Matcher mv = VIDEO_TYPE_PATTERN.matcher(video);
        if (mv.matches()) {
            bitrate = Integer.parseInt(mv.group(1));
        }
        if (duration < 16 * 60) {
            return bitrate == 2000;
        } else if (duration < 31 * 60) {
            return bitrate == 1000;
        } else if (duration < 61 * 60) {
            return bitrate == 600 || bitrate == 500;
        } else {
            return bitrate == 300;
        }
    }
    
    private void reportKuLockedBug() {
        try {
            NicoIdInfoCache.Entry info =
                    NicoIdInfoCache.getInstance().get(videoId.substring(2));
            Logger.warning("-----------------------------------");
            Logger.warning("GINZAkulowGWɃoO悤łD");
            Logger.warning("ȉ̃|[gNicoCacheXɂ񍐒܂ƍKłD");
            
            Logger.warning("WatchVars: duration=" + duration);
            Logger.warning("WatchVars: VIDEOS (all)");
            for (Map.Entry<String, Boolean> e : qualityVideos.entrySet()) {
                Boolean low = info == null ? null : info.dmcVideoEconomy.get(e.getKey());
                Logger.warning("WatchVars: - " + (e.getValue() ? "" : "!") + e.getKey()
                        + (low == null ? " null" : low ? " low" : "")
                        + (isKuLockedVideo(e.getKey()) ? " ku" : ""));
            }
            Logger.warning("WatchVars: VIDEOS (availables)");
            for (String video : videos) {
                Boolean low = info == null ? null : info.dmcVideoEconomy.get(video);
                Logger.warning("WatchVars: - " + video
                        + (low == null ? " null" : low ? " low" : ""));
            }
            Logger.warning("WatchVars: AUDIOS (all)");
            for (Map.Entry<String, Boolean> e : qualityAudios.entrySet()) {
                Boolean low = info == null ? null : info.dmcAudioEconomy.get(e.getKey());
                Logger.warning("WatchVars: - " + (e.getValue() ? "" : "!") + e.getKey()
                        + (low == null ? " null" : low ? " low" : "")
                        + (isKuLockedAudio(e.getKey()) ? " ku" : ""));
            }
            Logger.warning("WatchVars: AUDIOS (availables)");
            for (String audio : audios) {
                Boolean low = info == null ? null : info.dmcAudioEconomy.get(audio);
                Logger.warning("WatchVars: - " + audio
                        + (low == null ? " null" : low ? " low" : ""));
            }
            Logger.warning("-----------------------------------");
        } catch (Exception e) {
            Logger.warning("|[gɃG[܂D");
            Logger.error(e);
            Logger.warning("-----------------------------------");
        }
    }
    
    private boolean isKuLockedVideo(String video) {
        int bitrate = Integer.MAX_VALUE;
        String mode;
        
        Matcher m1 = VIDEO_TYPE_PATTERN.matcher(video);
        if (!m1.matches()) return false;
        
        bitrate = Integer.parseInt(m1.group(1));
        mode = m1.group(2);
        
        Matcher m2 = VIDEO_MODE_PATTERN.matcher(mode);
        if (!m2.matches()) return false;
        
        int height = Integer.parseInt(m2.group(1));
        
        // rbg[g2000𒴂铮̓|[go
        if (bitrate > 2000 && !Boolean.getBoolean("temporaryFlag2000kbps")) {
            Logger.info("*******************************");
            Logger.info("rbg[g2000kbps𒴂铮悪o܂:");
            Logger.info("  %s", video);
            Logger.info("  撷 = %d b", duration);
            Logger.info("temporaryFlag2000kbps=true ł̃|[g͕\ȂȂ܂D");
            Logger.info("*******************************");
        }
        
        if (duration < 16 * 60) {
            if (height <= 540) return bitrate > 1000;
            if (height <= 720) return bitrate > 2000;
            return true;
        } else if (duration < 31 * 60) {
            if (height <= 540) return bitrate > 1000;
            return true;
        } else if (duration < 61 * 60) {
            if (height <= 360) return bitrate > 600;
            return true;
        } else {
            if (height <= 360) return bitrate > 300;
            return true;
        }
    }
    
    private boolean isKuLockedAudio(String audio) {
        int bitrate = Integer.MAX_VALUE;
        
        Matcher m = AUDIO_TYPE_PATTERN.matcher(audio);
        if (!m.matches()) return false;
        
        bitrate = Integer.parseInt(m.group(1));
        
        return bitrate > 192;
    }
    
    private Map<String, Boolean> getQualityAvailability(JsonArray jsonQuality) {
        if (jsonQuality == null) {
            return null;
        }
        Map<String, Boolean> result = new LinkedHashMap<>();
        for (JsonValue o : jsonQuality.getList()) {
            if (o instanceof JsonObject) {
                JsonObject obj = (JsonObject)o;
                String id = obj.getString("id");
                Boolean available = obj.getBoolean("available");
                if (id != null && available != null) {
                    result.put(id, available);
                }
            }
        }
        return result;
    }
    
    private static String decodeURL(String value) {
        try {
            return URLDecoder.decode(value, "UTF-8");
        } catch (IllegalArgumentException e) {
            Logger.debug(e);
            return value;
        } catch (UnsupportedEncodingException e) {
            Logger.error(e);
        }
        return value;
    }
    
    private static String encodeURL(String value) {
        try {
            return URLEncoder.encode(value, "UTF-8");
        } catch (IllegalArgumentException e) {
            Logger.debug(e);
            return value;
        } catch (UnsupportedEncodingException e) {
            Logger.error(e);
        }
        return value;
    }
    
    /**
     * p[X JSON č\zĕԂ
     * @return č\z JSONAp[XłȂꍇ null
     */
    public String getJSON() {
        if (json != null) {
            return json.toJson();
        }
        return null;
    }
    
    /**
     * p[X JSON IuWFNgŕԂ
     * @return č\z JSONAp[XłȂꍇ null
     */
    public JsonObject getJsonObject() {
        return json;
    }
    
    /**
     * ID擾
     * @return IDA擾łȂꍇ null
     */
    public String getVideoId() {
        return videoId;
    }
    
    /**
     * ^Cg擾
     * @return ^CgA擾łȂꍇ null
     */
    public String getVideoTitle() {
        return videoTitle;
    }
    
    /**
     * v~AŃANZXĂ邩擾
     * @return v~AȂ true
     */
    public boolean isPremium() {
        return isPremium;
    }
    
    /**
     * 掿ɂȂ鎞ԑтɃANZXĂ邩擾
     * @return 掿ԑтȂ true
     */
    public boolean isPeakTime() {
        return isPeakTime;
    }
    
    /**
     * 悪V`ǂ擾
     * @return V`Ȃ true
     */
    public boolean isDmc() {
        return isDmc;
    }
    
    /**
     * p[X dmcInfo IuWFNgŕԂ
     * @return dmcInfoAp[XłȂꍇ null
     */
    public JsonObject getDmcInfo() {
        return dmcInfo;
    }
    
    /**
     * ^O擾
     * @return ^ÕXgA擾łȂꍇ͋󃊃Xg
     */
    public List<String> getTags() {
        ArrayList<String> tags = new ArrayList<String>();
        if (json != null) {
            JsonValue v;
            if (ginza) {
                v = json.get("videoDetail", "tagList");
            } else {
                v = json.get("tags");
            }
            if (v instanceof JsonArray) {
                for (JsonValue vv : ((JsonArray) v).getList()) {
                    String tag;
                    if (ginza) {
                        tag = vv.getString("tag");
                    } else {
                        tag = vv.getString("name");
                    }
                    if (tag != null) tags.add(tag);
                }
            }
        }
        return tags;
    }
    
    static HttpRequestHeader  REQUEST_HEADER_RIAPI;
    static HttpResponseHeader RESPONSE_HEADER_HTML;
    static HttpResponseHeader RESPONSE_HEADER_JSON;
    static HttpResponseHeader RESPONSE_HEADER_PLAIN;
    static {
        try {
            REQUEST_HEADER_RIAPI = newRequestHeader(
                    "http://riapi.nicovideo.jp/api/");
            RESPONSE_HEADER_HTML = new HttpResponseHeader(
                    "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\n\r\n");
            RESPONSE_HEADER_JSON = new HttpResponseHeader(
                    "HTTP/1.1 200 OK\r\nContent-Type: application/json\r\n\r\n");
            RESPONSE_HEADER_PLAIN = new HttpResponseHeader(
                    "HTTP/1.1 200 OK\r\nContent-Type: text/plain\r\n\r\n");
        } catch (IOException e) {
            Logger.error(e);
        }
    }
    
    private static HttpRequestHeader newRequestHeader(String url) {
        HttpRequestHeader requestHeader = null;
        try {
            requestHeader = new HttpRequestHeader(
                    HttpHeader.GET + " " + url + " HTTP/1.1\r\n\r\n");
            requestHeader.setMessageHeader("User-Agent",
                    TextUtil.escapeHTML(Main.VER_STRING));
        } catch (IOException e) {
            Logger.error(e);
        }
        return requestHeader;
    }
    
    /**
     * RXgN^ɓn watch y[W̓e擾
     * @return watch y[W̓e
     */
    public String getContent() {
        return content;
    }
    
    /**
     * watch y[W JSON ɑ΂ nlFilter Œue擾
     * <p>
     * ܂ JSON Ŝɑ΂ nlFilter KpŁAX JSON
     * f[^̓̒lɑ΂ăGXP[vĂ nlFilter
     * KpAăGXP[vsB
     * <p>
     * ΏۂƂȂ JSON ͈ȉ̒ʂ
     * <ul>
     * <li>watchAPI ƂĎ擾 JSON
     * <li>watch y[W watchAPIDataContainer  JSON
     * </ul>
     * <p>
     * uΏۂ̃f[^͈ȉ̒ʂ
     * <ul>
     * <li>videoDetail  description ̒l
     * </ul>
     * 
     * @return JSON u watch y[W̓eA
     * JSON ܂܂Ȃꍇ {@linkplain #getContent()} Ɠ
     */
    public String getReplacedContent() {
        if (!harajuku && videoId != null) {
            if (requestHeader == null) {
                requestHeader = newRequestHeader(
                        "http://www.nicovideo.jp/watch/" + videoId);
            }
            if (ginza) {
                replaceJSON("videoDetail", "description");
            } else {
                replaceJSON("video", "description");
            }
            
            String replaced = addCacheIcon(json.toJson());
            if (content != data && !rawJson) {
                StringBuilder sb = new StringBuilder(content);
                sb.replace(dataMatcher.start(1), dataMatcher.end(1),
                        TextUtil.escapeHTML(replaced));
                replaced = replacePlaylistData(sb);
            }
            return replaced;
        }
        return content;
    }
    
    // nlFilter g playlistData u
    private String replacePlaylistData(StringBuilder sb) {
        String replaced = sb.toString();
        Matcher m = PLAYLIST_DATA_CONTAINER_PATTERN.matcher(replaced);
        if (m.find()) {
            String playlistData = TextUtil.unescapeHTML(m.group(1));
            sb.replace(m.start(1), m.end(1),
                    TextUtil.escapeHTML(addCacheIcon(playlistData)));
            replaced = sb.toString();
        }
        return replaced;
    }
    
    // nlFilter g JSON ̒lu
    private void replaceJSON(String...args) {
        JsonValue v = this.getValue(args);
        if (v instanceof JsonString) {
            String s = v.toString();
            s = EasyRewriter.replace(null, s,
                    requestHeader, RESPONSE_HEADER_HTML);
            ((JsonString) v).setValue(s);
        }
    }
    
    // nlFilter găLbVt
    private String addCacheIcon(String data) {
        return EasyRewriter.replace(null, data,
                REQUEST_HEADER_RIAPI, RESPONSE_HEADER_JSON);
    }
    
    
    // WatchRewriteřĂяop(flvInfogetflv̒uKp)
    void replaceFlvInfo(String watch_url) {
        if (!ginza || videoId == null || flashvars == null) {
            return;
        }
        JsonValue v = flashvars.get("flvInfo");
        if (v instanceof JsonString) {
            String flvInfo = decodeURL(v.toString());
Logger.debug("original flvInfo: " + flvInfo);
            Matcher m = WATCH_ECONOMY_MODE_PATTERN.matcher(watch_url);
            String url_param = "/" + videoId + (m.find() ? m.group() : "");
            HttpRequestHeader getflvRequestHeader = newRequestHeader(
                    "http://flapi.nicovideo.jp/api/getflv" + url_param);
            flvInfo = EasyRewriter.replace(null,
                    GetFlvRewriter.replace(url_param, flvInfo),
                    getflvRequestHeader, RESPONSE_HEADER_PLAIN);
Logger.debug("replaced flvInfo: " + flvInfo);
            ((JsonString) v).setValue(encodeURL(flvInfo));
        }
    }

    private static final Pattern SM_FLV_PATTERN =
            Pattern.compile("^https?://(?!tn-skr)[^/]+(?:smilevideo|nicovideo)\\.jp/smile\\?(\\w)=([^.]+)\\.\\d+(as3)?(low)?$");
    void replaceDataApiData(String watch_url) {
        if (ginza || json == null || json.get("video") == null) {
            return;
        }

        String smid = getVideoId();
        
// low̃LbVuEUɎcĂ̑΍
// eco=1ŒʏLbVƂ͂̂܂܂ɂāAŗp
        Matcher m = WATCH_ECONOMY_MODE_PATTERN.matcher(watch_url);
        if (!m.find()) {
            String videoSource = json.getString("video", "smileInfo", "url");
            if (videoSource == null) {
                return;
            }
            
            String postfix = ".unknown";
            Matcher m_url = SM_FLV_PATTERN.matcher(videoSource);
            if (m_url.matches()) {
                String format = m_url.group(1);
                if (format.equals("v")) {
                    postfix = ".flv";
                } else if (format.equals("m")) {
                    postfix = ".mp4";
                } else if (format.equals("s")) {
                    postfix = ".swf";
                }
                
                // smidƓIDقȂꍇsmidŃLbV邽߂ɋLĂ
                String vid = m_url.group(2);
                String cid = videoId.substring(2);
                if (!vid.equals(cid) && !json.getBoolean("video", "isDeleted")) {
                    if (NLConfig.matches("deletedVideoId", vid)) {
                        // ȍ~͍폜ꂽƂď
                        json.getObject("video").put("isDeleted", new JsonTrue());
                    } else {
                        NLShared.INSTANCE.put_vid2cid(vid, cid);
                    }
                }
            }
            
            VideoDescriptor cachedVideo = Cache.getPreferredCachedVideo(smid, false, null);
            if (cachedVideo == null) { // not exists
                return;
            }
            Cache cache = new Cache(cachedVideo);
            if (videoSource.endsWith("low") && !cache.isLow()
                    && (Boolean.getBoolean("convertFlv2Mp4") || postfix.equals(cache.getPostfix()))) {
                String newSource = videoSource.substring(0, videoSource.length() - 3);
                Json.putString(newSource, json, "video", "smileInfo", "url");
            }
        }
    }
    
    /**
     * watch y[W JSON l擾ĕŕԂ
     * <p>
     * Zerȍꍇ watchAPIDataContainer  JSON ΏۂƂȂB
     * h̏ꍇ so.addVariable ̒lɍ\zAX var Video ̒l
     * videoDetail IuWFNgɃ}bv JSON ΏۂƂȂivCXg`jB
     * h̏ꍇ͓ŋ[I JSON Ƀ}bvĂ邽߈ȉ̈ႢB
     * <ul>
     * <li>Zero ł flashvars IuWFNg̒l JSON IuWFNgɂ
     * <li>Zero ł̓^Ö tagList Ahł tags/lockedTags ƂȂ
     * <li>Zero ɑ݂ uploaderInfo  viewerInfo ͖
     * <li>l^Ul̈ȅꍇƕ̏ꍇ
     * </ul>
     * <p>
     * JSON ͊Kw\Ȃ̂ŁAɂ͎擾Ώۂ̒l肷邽߂
     * o[Kw\ɕw肷î߉ψƂȂjB
     * Ⴆ΁AVideoDetail  description ̒l擾ꍇA
     * {@code getString("VideoDetail", "description")} ƌĂяoB
     * <p>
     * w肵ꍇA̒l݂Ȃꍇ flashvars Ď擾B
     * ̂߁AhɂΉɂ flashvars Œڎw肵ȂǂB
     * 
     * @param args 擾Ώۂ̒l肷郁o[
     * @return 擾Ώۂ̕A݂Ȃ or l̂k̏ꍇ null
     */
    @Deprecated
    public String getString(String...args) {
        if (json != null) {
            JsonValue v = getValue(args);
            if (v instanceof JsonNull) {
                return null;
            } else if (v != null) {
                return v.toString();
            }
        }
        return null;
    }
    
    private JsonValue getValue(String...args) {
        JsonValue v = json.get((Object[])args);
        if (v == null && args.length == 1 && flashvars != null) {
            return flashvars.get(args[0]);
        }
        return v;
    }
    
    /**
     * watch y[W JSON l擾Đ^UlŕԂ
     * @param args 擾Ώۂ̒l肷郁o[
     * @return 擾Ώۂ̒l݂^ƕ]Ƃ true
     * @see #getString(String...)
     */
    @Deprecated
    public boolean getBoolean(String...args) {
        if (json != null) {
            JsonValue v = getValue(args);
            if (v != null) {
                return v.toBoolean();
            }
        }
        return false;
    }
    
    /**
     * watch y[W JSON l擾Đl(int)ŕԂ
     * @param args 擾Ώۂ̒l肷郁o[
     * @return 擾Ώۂ̐lA݂Ȃ or ls or l̂[̏ꍇ 0
     * @see #getString(String...)
     */
    @Deprecated
    public int getInteger(String...args) {
        long l = getLong(args);
        return (int)l;
    }
    
    /**
     * watch y[W JSON l擾Đl(long)ŕԂ
     * @param args 擾Ώۂ̒l肷郁o[
     * @return 擾Ώۂ̐lA݂Ȃ or ls or l̂[̏ꍇ 0
     * @see #getString(String...)
     */
    @Deprecated
    public long getLong(String...args) {
        if (json != null) {
            JsonValue v = getValue(args);
            if (v instanceof JsonNumber) {
                return ((JsonNumber) v).getLong();
            } else if (v instanceof JsonString) {
                try {
                    return Long.parseLong(((JsonString) v).value());
                } catch (NumberFormatException e) {
                    Logger.debug(e);
                }
            }
        }
        return 0;
    }
    
    /**
     * watch y[W JSON l擾Ďl(double)ŕԂ
     * @param args 擾Ώۂ̒l肷郁o[
     * @return 擾Ώۂ̎lA݂Ȃ or ls or l̂[̏ꍇ 0
     * @see #getString(String...)
     */
    @Deprecated
    public double getDouble(String...args) {
        if (json != null) {
            JsonValue v = getValue(args);
            if (v instanceof JsonNumber) {
                return ((JsonNumber) v).getDouble();
            } else if (v instanceof JsonString) {
                try {
                    return Double.parseDouble(((JsonString) v).value());
                } catch (NumberFormatException e) {
                    Logger.debug(e);
                }
            }
        }
        return 0;
    }
    
    private static List<String> getArrayOfString(JsonArray a) {
        if (a != null) {
            List<JsonValue> al = a.getList();
            ArrayList<String> result = new ArrayList<String>(al.size());
            for (JsonValue v2 : al) {
                if (v2 instanceof JsonString) {
                    result.add(v2.toString());
                } else {
                    return null;
                }
            }
            return result;
        }
        return null;
    }
    
    // eXgp
    public static void main(String args[]) {
        for (int i = 0; i < args.length; i++) {
            FileReader fr = null;
            StringBuilder sb = new StringBuilder();
            try {
                fr = new FileReader(args[i]);
                int c;
                while ((c = fr.read()) != -1) {
                    sb.append((char) c);
                }
                fr.close();
                String input = TextUtil.unescapeHTML(sb.toString());
                System.out.println("INPUT = " + input);
                JsonValue json = Json.parse(input);
                if (json != null) {
                    System.out.println("PARSE = " + json.toJson());
                }
            } catch (IOException e) {
                System.out.println(e.toString());
            }
        }
    }
    
}
